@@ -1,4 +1,4 @@
-#!/usr/bin/python3
+#!/usr/bin/env python3
 # -*- coding: utf-8 -*-
 
 """
@@ -40,6 +40,7 @@ import sys
 import os
 import io
 import types
+import time
 import struct
 import select
 import re
@@ -54,6 +55,7 @@ TAG_HISTORY                     = 2
 TAG_COMPLETION                  = 3
 TAG_PROMPT                      = 4
 TAG_HOTKEY                      = 5
+TAG_SIGNAL                      = 6
 TAG_WHAT_ARE_YOUR_INTERESTS     = 127
 TAG_IGNORE                      = 251
 TAG_ADD_TO_COMPLETION_LIST      = 252
@@ -162,8 +164,8 @@ def write_patiently(fh, buffer):
             if (nwritten <= 0):
                 send_error("error writing: {0}".format(str(buffer)))
             already_written = already_written + nwritten
-        except BrokenPipeError: # ignore harmless error when rlwrap dies
-            pass
+        except BrokenPipeError: # quit when rlwrap dies
+            sys.exit(1)
 
 def read_message():
     """
@@ -231,6 +233,7 @@ def tag2name(tag):
                  'TAG_PROMPT',
                  'TAG_COMPLETION',
                  'TAG_HOTKEY',
+                 'TAG_SIGNAL',
                  'TAG_HISTORY',
                  'TAG_OUTPUT_OUT_OF_BAND',
                  'TAG_ERROR',
@@ -253,13 +256,16 @@ def send_warn(message):
     write_message(TAG_OUTPUT_OUT_OF_BAND, "{0}: {1}".format(__name__, message))
 
 
-def send_error(message,e):
+def send_error(message,e = None):
     """
     send message to rlwrap, and raise e.
     """
-    write_message(TAG_OUTPUT_OUT_OF_BAND, "{0}: {1}".format(__name__, message))
-    raise e
-
+    write_message(TAG_OUTPUT_OUT_OF_BAND if e else TAG_ERROR, "{0}: {1}".format(__name__, message))
+    if e:
+       raise e
+    else:
+       time.sleep(2) # timeout doesn't matter much because rlwrap will kill us anyway
+       exit()
 
 def intercept_error(func):
     """
@@ -277,38 +283,46 @@ def intercept_error(func):
     return wrapper
 
 
+
+def intercept_error_with_message(message=None):
+    """  
+    A decorator (-factory) to intercept an exception, send the message
+    to rlwrap, print a message and exit (or re-raise the exception, if
+    message = None) N.B: decorators, hence also <message> are evaluated
+    at compile time. @intercept_error_with_message (f"This script
+    crashed after {sec} seconds") doesn't make sense.
+    """
+    def intercept_error_closure(func):
+        def wrapper(*args, **kwargs):
+            try:
+                return func(*args, **kwargs)
+            except Exception as e:
+                complete_message = "{0}: {1}".format(__name__, '/'.join(map(str,e.args))) if message == None else "{0}\n{1}".format(message, e)
+                write_message(TAG_ERROR, complete_message)
+                if message:
+                    exit()
+                else:
+                    raise e
+        return wrapper
+    return intercept_error_closure
+
 def is_string(value):
-    """
-    A decorator to intercept an exception, send the message to rlwrap, and raise an exception.
-    """
     return isinstance(value, str) or value == None
 
 
 def is_boolean(value):
-    """
-    A decorator to intercept an exception, send the message to rlwrap, and raise an exception.
-    """
     return isinstance(value, bool) or value == None
 
 
 def is_integer(value):
-    """
-    A decorator to intercept an exception, send the message to rlwrap, and raise an exception.
-    """
     return isinstance(value, int) or value == None
 
 
 def is_float(value):
-    """
-    A decorator to intercept an exception, send the message to rlwrap, and raise an exception.
-    """
     return isinstance(value, numbers.Number) or value == None
 
 
 def is_callable(value):
-    """
-    A decorator to intercept an exception, send the message to rlwrap, and raise an exception.
-    """
     return isinstance(value, collections.Callable) or value == None
 
 
@@ -366,8 +380,7 @@ class RlwrapFilter:
             self.warn("{0} should not be '{1}'\n".format(name, type(value)))
 
         if name == 'minimal_rlwrap_version' and (value > rlwrap_version):
-            self.warn("requires rlwrap version {0} or newer.\n".format(str(value)))
-
+            self.error("requires rlwrap version {0} or newer.\n".format(str(value)))
         self.__dict__[name] = value
 
 
@@ -389,6 +402,7 @@ class RlwrapFilter:
             'output_handler':is_callable,
             'prompt_handler':is_callable,
             'hotkey_handler':is_callable,
+            'signal_handler':is_callable,
             'echo_handler':is_callable,
             'message_handler':is_callable,
             'history_handler':is_callable,
@@ -437,7 +451,7 @@ class RlwrapFilter:
             self.echo_has_been_handled = False
 
         if (not self.echo_has_been_handled):
-            if (re.match(r'\n', message)):
+            if (not re.search(r'\n', message)):
                 # save all output until we have one *whole* echo line
                 self.saved_output = self.saved_output + message
                 return ""
@@ -449,7 +463,7 @@ class RlwrapFilter:
                 (echo, nl, message) = re.match(r'^([^\n\r]*)(\r?\n)?(.*)?', message, re.DOTALL).groups()
                 handled_echo = when_defined(self.echo_handler, echo)
         self.cumulative_output = self.cumulative_output + message
-        return handled_echo + str(nl or "") + (when_defined(self.output_handler, message))
+        return handled_echo + str(nl or "") + str(when_defined(self.output_handler, message))
 
 
     def add_to_completion_list(self, *args):
@@ -497,12 +511,13 @@ class RlwrapFilter:
 
     def add_interests(self, message):
         interested = list(message)
-        tag2handler = {TAG_OUTPUT      : self.output_handler,
-                       TAG_INPUT       : self.input_handler,
+        tag2handler = {TAG_OUTPUT      : self.output_handler or self.echo_handler, # echo is the first OUTPUT after an INPUT
+                       TAG_INPUT       : self.input_handler or self.echo_handler,  # so to determine what is ECHO we need to see INPUT... 
                        TAG_HISTORY     : self.history_handler,
                        TAG_COMPLETION  : self.completion_handler,
                        TAG_PROMPT      : self.prompt_handler,
-                       TAG_HOTKEY      : self.hotkey_handler}
+                       TAG_HOTKEY      : self.hotkey_handler,
+                       TAG_SIGNAL      : self.signal_handler}
 
         for tag in range(0, len(message)):
             if interested[tag] == 'y':
@@ -530,7 +545,7 @@ class RlwrapFilter:
         send_warn(message)
 
 
-    def error(self, message,e):
+    def error(self, message,e = None):
         """
         send message to rlwrap, and raise e.
         """
@@ -544,7 +559,20 @@ class RlwrapFilter:
     def send_ignore_oob(self, text):
         write_message(TAG_IGNORE, text)
 
+      
+    def tweak_readline_oob(self, rl_tweak, *args):
+        nargs = {'rl_variable_bind'                   : 2,
+                 'rl_completer_word_break_characters' : 1,
+                 'rl_completer_quote_characters'      : 1,
+                 'rl_filename_completion_desired'     : 1}
+                 # the above list can be extended in future versions
+        if rl_tweak not in nargs: 
+            self.error("tweak_readline_oob() called with unknown/unimplemented readline function '{}'".format(rl_tweak))
+        if len(args) !=  nargs[rl_tweak]:
+            self.error("tweak_readline_oob({},...) should be called with exactly {} args".format(rl_tweak, nargs[rl_tweak] + 1))
+        self.send_ignore_oob("@" + "::".join((rl_tweak,) + args + ("\n",)))
 
+        
     def cwd(self):
         return os.getcwd()
 
@@ -556,9 +584,8 @@ class RlwrapFilter:
         # $RLWRAP_COMMAND_PID can be undefined (e.g. when run interactively, or under rlwrap -z listing
         # or == "0" (when rlwrap is called without a command name, like in rlwrap -z filter.py)
         # In both cases: print help text
-        if not os.environ.get('RLWRAP_COMMAND_PID'):
+        if os.environ.get('RLWRAP_COMMAND_PID') in [None, '0']:
             write_message(TAG_OUTPUT_OUT_OF_BAND, self.help_text + '\n')
-
         while(True):
             tag, message = read_message()
             
@@ -591,15 +618,17 @@ class RlwrapFilter:
                     write_message(tag,REJECT_PROMPT);
                     # don't update <previous_tag> and don't reset <cumulative_input>
                     next
-                if (os.environ.get('RLWRAP_IMPATIENT') and not re.match('\n$', self.cumulative_output)):
+                if (os.environ.get('RLWRAP_IMPATIENT') and not re.search('\n$', self.cumulative_output)):
                     # cumulative output contains prompt: chop it off!
                     # s/[^\n]*$// takes way too long on big strings,
                     # what is the optimal regex to do this?
                     self.cumulative_output = re.sub('(?<![^\n])[^\n]*$', '', self.cumulative_output)
 
                 response = when_defined(self.prompt_handler, message)
-                if (re.match('\n', response)):
+                if (re.search('\n', response)):
                     send_error('prompts may not contain newlines!')
+            elif (tag == TAG_SIGNAL):
+                response = when_defined(self.signal_handler, message)
             elif (tag == TAG_WHAT_ARE_YOUR_INTERESTS):
                 response = self.add_interests(message)
             else:
